//This function initializes server_fd variable

#include "webkit_client.h"
#include "ctxlog.h"
using namespace std;
int webkit_client::server_fd = -1;
PF* webkit_client::func_array;
std::mutex(webkit_client::mu);
std::condition_variable webkit_client::signal;
const int k_connect_timeout = 10;  // server connect timeout in seconds
const int k_next_connect_wait = 1;  // next connect wait interval in seconds

//Close the socket, close more gracefully by sending cancel event to g_recv_thread_id
void webkit_client::deinitComm(){
    int err;
    CtxTraceInfo(TC_WEBKIT,"+");


    if(server_fd != -1){
        err = close(server_fd);
        if(err == -1){
            std::cerr<<"Could not close webkit_client::server_fd, error: "<<strerror(errno)<<std::endl;
            CtxTraceError(TC_WEBKIT,"Could not close server_fd, error: %s\n", strerror(errno));
        }
        server_fd = -1;
    }
    else{
        CtxTraceWarning(TC_WEBKIT,"server_fd is already closed\n");
    }

}


void webkit_client::initComm(PF* arr, char* sock){

    pthread_attr_t tattr;
    CtxTraceInfo(TC_WEBKIT,"+");
    //First inititalize func_array
    func_array = arr;
    pthread_t recv_thread_id; //Thread that runs recvMessage
   

    pthread_attr_init(&tattr);
    pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
    //Create a thtread. This thread will initialize server_fd and start receiveing messages in an infinite loop
    pthread_create(&recv_thread_id, &tattr, sinitComm, (void*)sock);
    pthread_attr_destroy(&tattr);

}

//Connect to server_fd and start receiveing messages.
void* webkit_client::sinitComm(void* args){


    char* socket_path = (char*)args;
    struct sockaddr_un addr;
    int connect_result = -1, timeout = k_connect_timeout;
    std::unique_lock<std::mutex> lock(webkit_client::mu);  // lock until socket is initialized and message is sent
    if ( (server_fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
        std::cerr<<"Could not create socket!"<<std::endl;
        CtxTraceError(TC_WEBKIT,"Could not create socket");
        pthread_exit(NULL);
    }

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, socket_path, sizeof(addr.sun_path)-1);

    connect_result = connect(server_fd, (struct sockaddr*)&addr, sizeof(addr));  // Try first time connect without wait
    // below loop get tries till maxtimeout
    while (timeout > 0 && connect_result == -1) {
        sleep(k_next_connect_wait);
        connect_result = connect(server_fd, (struct sockaddr*)&addr, sizeof(addr));
        timeout = timeout - k_next_connect_wait;
        CtxTraceInfo(TC_WEBKIT, "Failed to connect to server, trying again after %d secs, retries left : %d",
                     k_next_connect_wait, timeout);
    }

    //Exit if connect does not happens after max retries.
    if ( connect_result == -1) {
        CtxTraceError(TC_WEBKIT,"Connection error with webkit_server");
        std::cerr<<"Connection error!"<<std::endl;
        pthread_exit(NULL);
    }

    /* Now as the connection is established, we can move ahead and send the PID of the extension process
        This is the best place to do it */
    std::vector<int> intList;
    intList.push_back(getpid());
    std::vector<std::string> stringList;
    webkit_message *s = webkit_util::wrapMessage(FUNC_PID_EXT, intList, stringList);
    webkit_client::sendMessage(s);
    free(s);

    free(socket_path);

    //send signal to unlock via condition_variable 
    lock.unlock();
    webkit_client::signal.notify_one();

    //recvMessage indefinitely
    webkit_client::recvMessage();


    return (void*)0;
}



void webkit_client::processMessage(webkit_message* &s){

    int args1[s->total_int_args];
    //Here adding extra '/0' string in the end so that ctxs_cb_wrapper can calculate the total length of the array.
    std::string args2[s->total_string_args + 1]; 

    int cur = 0;
    char* buf = (char*)malloc(BUFSIZE);

    for(int i=0; i<s->total_int_args; i++){

        args1[i] = webkit_util::parseNextInt(cur, s->serialized_arg, buf);

    }

    for(int i=0;i<s->total_string_args;i++){

        args2[i] = webkit_util::parseNextString(cur, s->serialized_arg, buf);

    }
    args2[s->total_string_args] = "\0";


    func_array[s->func_name](args1, args2);

    free(buf);

}

//infinite loop running function which waits for any data on the server_fd
void webkit_client::recvMessage(){

    webkit_message *s;
    int wrv;
    char* buf1 = (char*)malloc(BUFSIZE);
    if(buf1 == NULL){
        std::cerr<<"webkit_client::recvMessage could not malloc, failing"<<std::endl;
        CtxTraceError(TC_WEBKIT, "could not malloc %d, too less memory?", BUFSIZE);
        pthread_exit(NULL);
    }
    int total_len_of_string;

    if(server_fd == -1){
        std::cerr<<"webkit_client::recvMessage Error in server_fd"<<std::endl;
        CtxTraceError(TC_WEBKIT,"server_fd = -1, haven't connected with server yet");
        pthread_exit(NULL);
    }

    memset(buf1, 0, BUFSIZE);

    while(1){



        while(1){
            wrv = read(server_fd, buf1, 4*sizeof(int));
            if(wrv==0){
             continue;
            }
            else{
                break;
            }
                
        }
        CtxTraceVerbose(TC_WEBKIT, "Value of server_fd %d", server_fd);

        if(wrv != 4*sizeof(int)){

            std::cerr<<"webkit_client::recvMessage Error in reading first four ints in recvMessage\
                        in webkit_client "<<strerror(errno)<<std::endl;

            CtxTraceFatal(TC_WEBKIT,"Error in reading first four ints: %s\n", strerror(errno));
            pthread_exit(NULL);

        }

        total_len_of_string = *(int *)(buf1 + 3*sizeof(int));

        s = (webkit_message*)malloc(sizeof(webkit_message) + total_len_of_string);

        s->func_name = *(int*)buf1;
        s->total_int_args = *((int *)((char*)buf1 + 1*sizeof(int)));
        s->total_string_args = *((int *)((char*)buf1 + 2*sizeof(int)));
        s->total_len_of_string = *(int *)((char*)buf1 + 3*sizeof(int));

        if(s->total_len_of_string > BUFSIZE){
            CtxTraceFatal(TC_WEBKIT,"BUFSIZE VALUE: %d", BUFSIZE);
            CtxTraceFatal(TC_WEBKIT, "s->total_len_of_string is more than the capacity BUFSIZE VALUE, exiting thread \n");
            pthread_exit(NULL);
        }

        memset(buf1, 0, BUFSIZE);
        int to_be_read = s->total_len_of_string;

        if(s->total_len_of_string != 0){

            while(to_be_read != 0){

                wrv = read(server_fd, buf1, to_be_read);
                if(wrv==0){
                    continue;
                }
                else{
                    if(wrv < 0){
                        std::cerr<<"webkit_client::recvMessage Error in\
                                    reading in webkit_client: "<<strerror(errno)<<std::endl;
                        CtxTraceError(TC_WEBKIT, "Error in reading in webkit_client, exiting thread %s", strerror(errno));
                        pthread_exit(NULL);

                    }
                    else{
                        memcpy((s->serialized_arg) + (s->total_len_of_string - to_be_read), buf1, wrv);
                        to_be_read = to_be_read - wrv;
                        assert(to_be_read >=0);
                        memset(buf1, 0, BUFSIZE);
                    }

                }
                    
            }
        }
        processMessage(s);
        free(s);
    }

    free(buf1);
}



void webkit_client::sendMessage(webkit_message* &s){

    int wrv;
    int size_struct = sizeof(webkit_message) + s->total_len_of_string;


    
    if(server_fd == -1){
        std::cerr<<"webkit_client::sendMessage server_fd not initialized yet"<<std::endl;
        CtxTraceError(TC_WEBKIT,"server_fd not initialized yet, returning");
        return;
    }

    wrv = write(server_fd, s, size_struct);
    if(wrv != size_struct){
        if(wrv > 0){
            CtxTraceFatal(TC_WEBKIT, "Partial write, exiting, value of wrv is %d\n", wrv);
            CtxLogShutdown();
            exit(-1);
        }
        else{
            CtxTraceFatal(TC_WEBKIT,"Could not write, error: %s, exiting\n", strerror(errno));
            CtxLogShutdown();
            std::cerr<<"Error writing into the socket"<<std::endl;
            exit(-1);
        }
    }
}

